#include "PCSPE.h"
#include "GUI.h"



AudioEffect* createEffectInstance(audioMasterCallback audioMaster)
{
	return new PCSPE(audioMaster);
}



PCSPE::PCSPE(audioMasterCallback audioMaster) : AudioEffectX(audioMaster, NUM_PROGRAMS, NUM_PARAMS)
{
	int row,pgm,chn;

	setEditor((AEffEditor*)new GUI(this));

	setNumInputs(NUM_INPUTS);
	setNumOutputs(NUM_OUTPUTS);
	setUniqueID(PLUGIN_UID);

	isSynth();
	canProcessReplacing();
	programsAreChunks(true);

	MidiChannel=0;

	for(chn=0;chn<16;++chn)
	{
		Program[chn]=0;
		MultiMode[chn]=false;

		MidiPitchBend[chn]=0;
		MidiPitchBendRange[chn]=2.0f;
	}

	for(pgm=0;pgm<NUM_PROGRAMS;++pgm)
	{
		sprintf(ProgramName[pgm],"%3.3i default",pgm+1);

		memset(&EnvelopeV[pgm],0,ENVELOPE_TEXT_LEN);
		memset(&EnvelopeA[pgm],0,ENVELOPE_TEXT_LEN);
		memset(&EnvelopeP[pgm],0,ENVELOPE_TEXT_LEN);

		strcpy(EnvelopeV[pgm],"2");
		strcpy(EnvelopeA[pgm],"0");
		strcpy(EnvelopeP[pgm],"0");
	}

	pgm=0;
	row=0;

	while(1)
	{
		if(!strlen(PresetData[row])) break;

		sprintf(ProgramName[pgm],"%3.3i %s",pgm+1,PresetData[row++]);

		strcpy(EnvelopeV  [pgm],PresetData[row++]);
		strcpy(EnvelopeA  [pgm],PresetData[row++]);
		strcpy(EnvelopeP  [pgm],PresetData[row++]);

		++pgm;
	}

	CompileInstruments();

	SetUpdateGUI();

	for(chn=0;chn<SYNTH_CHANNELS;++chn)
	{
		SynthChannel[chn].note=-1;

		SynthChannelOrder[chn]=chn;
	}

	FrameRate=60;

	pOutputGain=.5f;

	LogActive=false;
	LogAuto=false;
	LogAutoEnabled=false;
	strcpy(LogFileName,"");

	sFrameAcc=0;

	sPITEnable=false;
	sPITOutput=0;
	sPITCount=0;
	sPITReload=1193180/440;
	sPITResampleAcc=0;

	MidiQueue.clear();

	MidiRPNLSB=0;
	MidiRPNMSB=0;
	MidiDataLSB=0;
	MidiDataMSB=0;

	memset(MidiKeyState,0,sizeof(MidiKeyState));

	suspend();
}



PCSPE::~PCSPE()
{
	LogStop();
}



void PCSPE::setParameter(VstInt32 index,float value)
{
}



float PCSPE::getParameter(VstInt32 index)
{
	return 0;
} 



void PCSPE::getParameterName(VstInt32 index,char *label)
{
} 



void PCSPE::getParameterDisplay(VstInt32 index,char *text)
{
} 



void PCSPE::getParameterLabel(VstInt32 index,char *label)
{
} 



VstInt32 PCSPE::getProgram(void)
{
	return Program[MidiChannel];
}



void PCSPE::setProgram(VstInt32 program)
{
	if(Program[MidiChannel]!=program)
	{
		Program[MidiChannel]=program;
	}
}



void PCSPE::getProgramName(char* name)
{
	strcpy(name,ProgramName[Program[MidiChannel]]); 
}



void PCSPE::setProgramName(char* name)
{
	strncpy(ProgramName[Program[MidiChannel]],name,MAX_NAME_LEN);

	ProgramName[Program[MidiChannel]][MAX_NAME_LEN-1]='\0';
} 



VstInt32 PCSPE::canDo(char* text)
{
	if(!strcmp(text,"receiveVstEvents"   )) return 1;
	if(!strcmp(text,"receiveVstMidiEvent")) return 1;

	return -1;
}



VstInt32 PCSPE::getNumMidiInputChannels(void)
{
	return 16;
}



VstInt32 PCSPE::getNumMidiOutputChannels(void)
{
	return 0;//no MIDI output
}



VstInt32 PCSPE::SaveStringChunk(char *str,float *dst)
{
	unsigned int i;

	for(i=0;i<strlen(str);++i) *dst++=str[i];

	*dst++=0;

	return strlen(str)+1;
}



VstInt32 PCSPE::LoadStringChunk(char *str,int max_length,float *src)
{
	unsigned int c,ptr;

	ptr=0;

	while(1)
	{
		c=(unsigned int)*src++;
		
		str[ptr++]=c;

		if(!c) break;

		if(ptr==max_length-1)
		{
			str[ptr]=0;
			break;
		}
	}

	return ptr;
}



VstInt32 PCSPE::SavePresetChunk(float *chunk)
{
	int chn,ptr,pgm;

	ptr=0;

	chunk[ptr++]=DATA;

	chunk[ptr++]=MPRG; for(chn=0;chn<16;++chn) chunk[ptr++]=(float)Program[chn];

	for(pgm=0;pgm<NUM_PROGRAMS;++pgm)
	{
		chunk[ptr++]=PROG; chunk[ptr++]=(float)pgm;

		chunk[ptr++]=NAME; ptr+=SaveStringChunk(ProgramName[pgm],&chunk[ptr]);

		chunk[ptr++]=EVOL; ptr+=SaveStringChunk(EnvelopeV[pgm],&chunk[ptr]);
		chunk[ptr++]=EARP; ptr+=SaveStringChunk(EnvelopeA[pgm],&chunk[ptr]);
		chunk[ptr++]=EPTC; ptr+=SaveStringChunk(EnvelopeP[pgm],&chunk[ptr]);
	}

	chunk[ptr++]=FRMR; chunk[ptr++]=FrameRate;

	chunk[ptr++]=MMOD; for(chn=0;chn<16;++chn) chunk[ptr++]=MultiMode[chn]?1.0f:0;

	chunk[ptr++]=GAIN; chunk[ptr++]=pOutputGain;
	chunk[ptr++]=DONE;

	return ptr*sizeof(float);//size in bytes
}



void PCSPE::LoadPresetChunk(float *chunk)
{
	int chn,pgm;
	float tag;

	while(1)
	{
		tag=*chunk++;

		if(tag==DONE) break;

		if(tag==MPRG) for(chn=0;chn<16;++chn) Program[chn]=(int)*chunk++;

		if(tag==PROG) pgm=(int)*chunk++;

		if(tag==NAME) chunk+=LoadStringChunk(ProgramName[pgm],MAX_NAME_LEN,chunk);

		if(tag==EVOL) chunk+=LoadStringChunk(EnvelopeV[pgm],ENVELOPE_TEXT_LEN,chunk);
		if(tag==EARP) chunk+=LoadStringChunk(EnvelopeA[pgm],ENVELOPE_TEXT_LEN,chunk);
		if(tag==EPTC) chunk+=LoadStringChunk(EnvelopeP[pgm],ENVELOPE_TEXT_LEN,chunk);

		if(tag==FRMR) FrameRate=*chunk++;

		if(tag==MMOD) for(chn=0;chn<16;++chn) MultiMode[chn]=*chunk++?true:false;

		if(tag==GAIN) pOutputGain=*chunk++;
	}

	CompileInstruments();
}



VstInt32 PCSPE::getChunk(void** data,bool isPreset)
{
	int size;

	size=SavePresetChunk(Chunk);

	*data=Chunk;

	return size;
}



VstInt32 PCSPE::setChunk(void* data,VstInt32 byteSize,bool isPreset)
{
	if(byteSize>sizeof(Chunk)) return 0;

	memcpy(Chunk,data,byteSize);

	LoadPresetChunk(Chunk);

	SetUpdateGUI();

	return 0;
}



void PCSPE::SetUpdateGUI(void)
{
	UpdateGuiFlag=true;
}



void PCSPE::UpdateGUI(bool display)
{
	if(editor) ((GUI*)editor)->Update();

	if(display) updateDisplay();
}



const char* PCSPE::CompileEnvelope(char *str,envelopeStruct *env,bool hold)
{
	int ptr,len,loop,repeat_len,repeat_cnt;
	int c,num,num_prev,prefix,time_scale;
	char *err;

	env->len=0;
	env->loop_ptr=0;
	env->has_loop=false;
	env->error=true;	//error by default, cleared on successful parsing

	err=NULL;

	len=strlen(str);
	ptr=0;

	loop=-1;			//loop not defined yet
	num_prev=0;			//number to repeat
	time_scale=1;
	repeat_cnt=0;
	repeat_len=0;

	while(ptr<len)
	{
		while(str[ptr]==' ') ++ptr;	//skip spaces

		c=str[ptr];

		if(!c) break;

		if(c>='a'&&c<='z') c-=32;	//case insensitive

		prefix=PARSE_PREFIX_NONE;

		if(c=='-') prefix=PARSE_PREFIX_NEGATIVE;
		if(c=='*') prefix=PARSE_PREFIX_REPEAT;
		if(c=='X') prefix=PARSE_PREFIX_TIME_SCALE;
		if(c=='#') prefix=PARSE_PREFIX_ABSOLUTE;

		if(c=='|'||c=='L')	//loop marker
		{
			if(loop<0)
			{
				loop=env->len;
			}
			else
			{
				err="Loop point redefinition";
				break;
			}

			++ptr;
		}
		else
		if(prefix>PARSE_PREFIX_NONE||(c>='0'&&c<='9'))	//parse number with an optional prefix
		{
			if(prefix>PARSE_PREFIX_NONE)
			{
				++ptr;

				c=str[ptr];

				if(!(c>='0'&&c<='9'))
				{
					err="Prefix is not followed by a number";
					break;
				}
			}

			num=0;

			while(ptr<len)
			{
				c=str[ptr];

				if(c>='a'&&c<='z') c-=32;

				if(c<=32||c=='|'||c=='L'||c=='*') break;

				++ptr;

				if(c>='0'&&c<='9')
				{
					num=num*10+(c-'0');
					continue;
				}

				err="Unexpected character while parsing a number";
				break;
			}

			if(prefix==PARSE_PREFIX_TIME_SCALE)
			{
				time_scale=num;

				if(time_scale<1) time_scale=1;

				continue;
			}

			if(prefix==PARSE_PREFIX_NEGATIVE) num=-num;

			if(prefix==PARSE_PREFIX_REPEAT)
			{
				++repeat_cnt;

				repeat_len=num-1;
		
				num=num_prev;
			}
			else
			{
				repeat_cnt=0;

				num_prev=num;
			}

			if(repeat_cnt>1)
			{
				err="More than one repeat";
				break;
			}

			if(repeat_len<2) repeat_len=1;

			repeat_len*=time_scale;

			if(num<-ENVELOPE_NUM_RANGE) num=-ENVELOPE_NUM_RANGE;
			if(num> ENVELOPE_NUM_RANGE) num= ENVELOPE_NUM_RANGE;

			while(repeat_len--)
			{
				if(prefix==PARSE_PREFIX_ABSOLUTE) env->abs[env->len]=1;

				env->value[env->len++]=num;
			}

			if(env->len>ENVELOPE_DATA_LEN)
			{
				err="Compiled envelope data is too long";
				break;
			}
		}
		else
		{
			err="Unexpected character while parsing string";
			break;
		}
	}

	if(!err&&env->len)	//compiled successfully
	{
		env->error=false;

		if(loop<0)
		{
			if(!hold)
			{
				env->value[env->len++]=0;	//needed for pitch envelope to prevent accumulation without loop

				if(env->len>ENVELOPE_DATA_LEN)
				{
					err="Compiled envelope data is too long";
				}
			}

			env->loop_ptr=env->len-1;	//loop to last position, but loop flag is not set
		}
		else
		{
			env->has_loop=true;
			env->loop_ptr=loop;
		}

		if(env->loop_ptr>env->len-1) env->loop_ptr=env->len-1;
		if(env->loop_ptr<0) env->loop_ptr=0;
	}

	if(err||!env->len)	//make default envelope if error or empty line
	{
		env->len=1;
		env->loop_ptr=0;
		env->value[0]=0;
	}

	return err;
}



void PCSPE::CompileInstruments(void)
{
	int pgm;

	for(pgm=0;pgm<NUM_PROGRAMS;++pgm)
	{
		CompileEnvelope(EnvelopeV[pgm],&Instrument[pgm].EnvVol,true);
		CompileEnvelope(EnvelopeA[pgm],&Instrument[pgm].EnvArp,true);
		CompileEnvelope(EnvelopeP[pgm],&Instrument[pgm].EnvPitch,false);
	}
}



void PCSPE::LogStart(void)
{
	char name[1024];
	int i;

	if(LogActive) return;

	LogFile=fopen(LogFileName,"wt");

	switch(LogFormat)
	{
	case LOG_FORMAT_CSV:
		{
			fprintf(LogFile,"%1.1f\n",FrameRate);
		}
		break;

	case LOG_FORMAT_QB:
		{
			fprintf(LogFile,"REM FRAMERATE %1.1f\n\n",FrameRate);
		}
		break;

	case LOG_FORMAT_H:
		{
			strcpy(name,LogFileName);

			for(i=strlen(name)-1;i>=0;--i)
			{
				if(name[i]=='.') name[i]=0;
				if(name[i]=='\\'||name[i]=='/') { ++i; break; }
			}

			strcpy(name,&name[i]);

			fprintf(LogFile,"const unsigned char %s_data[]={\n",name);
			fprintf(LogFile,"0x%2.2x,//frame rate\n",(int)FrameRate);
		}
		break;
	}
	
	LogFrameDelta=0;
	LogDivPrev=-1;
	LogActive=true;
	LogAutoEnabled=false;

	SetUpdateGUI();
}



void PCSPE::LogStop(void)
{
	if(!LogActive) return;

	LogFrameFlush();

	switch(LogFormat)
	{
	case LOG_FORMAT_CSV:
		{
		}
		break;

	case LOG_FORMAT_QB:
		{
			fprintf(LogFile,"\nREM END\n");
		}
		break;

	case LOG_FORMAT_H:
		{
			fprintf(LogFile,"0x00\n};\n");
		}
		break;
	}

	fclose(LogFile);

	LogActive=false;
	LogAutoEnabled=false;

	SetUpdateGUI();
}



void PCSPE::LogFrameFlush(void)
{
	if(LogDivPrev>=0&&LogFrameDelta>0)
	{
		switch(LogFormat)
		{
		case LOG_FORMAT_CSV:
			{
				fprintf(LogFile,"%i,%i\n",LogFrameDelta,LogDivPrev);
			}
			break;

		case LOG_FORMAT_QB:
			{
				fprintf(LogFile,"DATA %i,%i\n",LogFrameDelta,LogDivPrev);
			}
			break;

		case LOG_FORMAT_H:
			{
				fprintf(LogFile,"0x%2.2x,0x%2.2x,0x%2.2x,\n",LogFrameDelta,LogDivPrev&255,LogDivPrev>>8);
			}
			break;
		}
	}
}



void PCSPE::LogFrame(void)
{
	VstInt32 n;

	if(!LogActive) return;

	if(LogAuto&&!LogAutoEnabled) return;

	n=sPITReload;

	if(!sPITEnable) n=0;

	if(LogDivPrev!=n)
	{
		LogFrameFlush();

		LogDivPrev=n;
		LogFrameDelta=0;
	}

	++LogFrameDelta;
}



void PCSPE::MidiAddNote(VstInt32 delta,VstInt32 channel,VstInt32 note,VstInt32 velocity)
{
	MidiQueueStruct entry;

	entry.type    =mEventTypeNote;
	entry.channel =channel;
	entry.delta   =delta;
	entry.note    =note;
	entry.velocity=velocity;//0 for key off

	MidiQueue.push_back(entry);
}



void PCSPE::MidiAddPitchBend(VstInt32 delta,VstInt32 channel,float depth)
{
	MidiQueueStruct entry;

	entry.type   =mEventTypePitchBend;
	entry.channel=channel;
	entry.delta  =delta;
	entry.depth  =depth;

	MidiQueue.push_back(entry);
}



bool PCSPE::MidiIsAnyKeyDown(void)
{
	int i;

	for(i=0;i<128;++i) if(MidiKeyState[i]) return true;

	return false;
}



void PCSPE::MidiAddProgramChange(VstInt32 delta,VstInt32 channel,VstInt32 program)
{
	MidiQueueStruct entry;

	entry.type   =mEventTypeProgram;
	entry.channel=channel;
	entry.delta  =delta;
	entry.program=program;

	MidiQueue.push_back(entry);
}



void PCSPE::MidiAddControl(VstInt32 delta,VstInt32 channel,VstInt32 value)
{
	MidiQueueStruct entry;

	entry.type   =mEventTypeControl;
	entry.channel=channel;
	entry.delta  =delta;
	entry.value  =value;

	MidiQueue.push_back(entry);
}



VstInt32 PCSPE::processEvents(VstEvents* ev) 
{ 
	VstMidiEvent* event;
	VstInt32 i,channel,status,note,velocity,wheel;
	char* midiData;

	for(i=0;i<ev->numEvents;++i) 
	{ 
		event=(VstMidiEvent*)ev->events[i];

		if(event->type!=kVstMidiType) continue; 

		midiData=event->midiData;

		channel=midiData[0]&0x0f;
		status =midiData[0]&0xf0;

		switch(status)
		{
		case 0x80:	//note off
		case 0x90:	//note on
			{
				note=midiData[1]&0x7f; 

				velocity=midiData[2]&0x7f; 

				if(status==0x80) velocity=0; 

				MidiAddNote(event->deltaFrames,channel,note,velocity);
			} 
			break;

		case 0xb0:	//control change
			{ 
				if(midiData[1]==0x64) MidiRPNLSB =midiData[2]&0x7f;
				if(midiData[1]==0x65) MidiRPNMSB =midiData[2]&0x7f;
				if(midiData[1]==0x26) MidiDataLSB=midiData[2]&0x7f;

				if(midiData[1]==0x06)
				{
					MidiDataMSB=midiData[2]&0x7f;

					if(MidiRPNLSB==0&&MidiRPNMSB==0) MidiPitchBendRange[channel]=(float)MidiDataMSB*.5f;
				}

				if(midiData[1]==0x14||midiData[1]==0x15)//unused controllers 20 and 21
				{
					MidiAddControl(event->deltaFrames,channel,midiData[1]);//remember controller
				}

				if(midiData[1]>=0x7b)//all notes off and mono/poly mode changes that also requires to do all notes off
				{
					MidiAddNote(event->deltaFrames,channel,-1,0);
				}
			} 
			break;

		case 0xc0:	//program change
			{
				MidiAddProgramChange(event->deltaFrames,channel,midiData[1]&0x7f);
			}
			break;

		case 0xe0:	//pitch bend change
			{
				wheel=(midiData[1]&0x7f)|((midiData[2]&0x7f)<<7);

				MidiAddPitchBend(event->deltaFrames,channel,(float)(wheel-0x2000)*MidiPitchBendRange[channel]/8192.0f);
			}
			break;
		}
	} 

	return 1; 
} 



VstInt32 PCSPE::SynthAllocateVoice(VstInt32 midi_ch,VstInt32 note)
{
	VstInt32 chn;

	for(chn=0;chn<SYNTH_CHANNELS;++chn) if(SynthChannel[chn].midi_ch==midi_ch&&SynthChannel[chn].note==note) return chn;
	for(chn=0;chn<SYNTH_CHANNELS;++chn) if(SynthChannel[chn].note<0) return chn;

	return -1;
}



void PCSPE::SynthChannelChangeNote(VstInt32 chn,VstInt32 midi_ch,VstInt32 note)
{
	VstInt32 ins;

	ins=Program[midi_ch];

	if(MultiMode[midi_ch])
	{
		ins+=(note-60);

		while(ins<0) ins+=NUM_PROGRAMS;
		while(ins>=NUM_PROGRAMS) ins-=NUM_PROGRAMS;

		note=60;
	}

	SynthChannel[chn].note=note;
	SynthChannel[chn].midi_ch=midi_ch;

	SynthChannel[chn].ptr_vol=0;
	SynthChannel[chn].ptr_arp=0;
	SynthChannel[chn].ptr_pitch=0;

	SynthChannel[chn].pitch_acc=0;
	SynthChannel[chn].arp_offset=0;
	SynthChannel[chn].vol_output=0;

	SynthChannel[chn].env_vol  =&Instrument[ins].EnvVol;
	SynthChannel[chn].env_arp  =&Instrument[ins].EnvArp;
	SynthChannel[chn].env_pitch=&Instrument[ins].EnvPitch;
}



void PCSPE::SynthChannelReleaseNote(VstInt32 midi_ch,VstInt32 note)
{
	int chn;

	if(note>=0)
	{
		for(chn=0;chn<SYNTH_CHANNELS;++chn)
		{
			if(SynthChannel[chn].midi_ch==midi_ch&&SynthChannel[chn].note==note) SynthChannel[chn].note=-1;
		}
	}
	else
	{
		for(chn=0;chn<SYNTH_CHANNELS;++chn)
		{
			if(SynthChannel[chn].midi_ch==midi_ch||midi_ch<0) SynthChannel[chn].note=-1;
		}
	}
}



void PCSPE::processReplacing(float **inputs,float **outputs,VstInt32 sampleFrames)
{
	float *outL=outputs[0];
	float *outR=outputs[1];
	float level,loops,freq,sample_rate;
	unsigned int i;
	int chn,add,vol_max;

	sample_rate=(float)updateSampleRate();

	while(--sampleFrames>=0)
	{
		for(i=0;i<MidiQueue.size();++i)
		{
			if(MidiQueue[i].delta>=0)
			{
				--MidiQueue[i].delta;

				if(MidiQueue[i].delta<0)//new message
				{
					switch(MidiQueue[i].type)
					{
					case mEventTypeNote:
						{
							if(MidiQueue[i].velocity)//key on
							{
								MidiKeyState[MidiQueue[i].note]=1;

								chn=SynthAllocateVoice(MidiQueue[i].channel,MidiQueue[i].note);

								if(chn>=0)
								{
									SynthChannelChangeNote(chn,MidiQueue[i].channel,MidiQueue[i].note);
								}
							}
							else//key off
							{
								if(MidiQueue[i].note>=0)
								{
									MidiKeyState[MidiQueue[i].note]=0;

									SynthChannelReleaseNote(MidiQueue[i].channel,MidiQueue[i].note);
								}
								else
								{
									memset(MidiKeyState,0,sizeof(MidiKeyState));

									SynthChannelReleaseNote(-1,-1);
								}
							}
						}
						break;

					case mEventTypePitchBend:
						{
							MidiPitchBend[MidiQueue[i].channel]=MidiQueue[i].depth;
						}
						break;

					case mEventTypeProgram:
						{
							if(Program[MidiQueue[i].channel]!=MidiQueue[i].program)
							{
								Program[MidiQueue[i].channel]=MidiQueue[i].program;

								SetUpdateGUI();
							}
						}
						break;

					case mEventTypeControl:
						{
							if(LogActive&&LogAuto)
							{
								if(MidiQueue[i].value==0x14)
								{
									if(!LogAutoEnabled)
									{
										LogAutoEnabled=true;
									}
								}
								else
								if(MidiQueue[i].value==0x15)
								{
									if(LogAutoEnabled)
									{
										LogAutoEnabled=false;

										LogStop();
									}
								}
							}
								
						}
						break;
					}
				}
			}
		}

		//advance envelopes and update counter

		sFrameAcc+=1.0f/(sample_rate/FrameRate);

		while(sFrameAcc>=1.0f)
		{
			sFrameAcc-=1.0f;

			for(chn=0;chn<SYNTH_CHANNELS;++chn)
			{
				if(SynthChannel[chn].note<0) continue;

				if(SynthChannel[chn].ptr_vol>=SynthChannel[chn].env_vol->len)
				{
					if(!SynthChannel[chn].vol_output&&!SynthChannel[chn].env_vol->has_loop) //terminate channel that is looped to zero volume at end
					{
						SynthChannel[chn].note=-1;
						continue;
					}

					SynthChannel[chn].ptr_vol=SynthChannel[chn].env_vol->loop_ptr;
				}

				SynthChannel[chn].vol_output=SynthChannel[chn].env_vol->value[SynthChannel[chn].ptr_vol++];

				if(SynthChannel[chn].ptr_arp>=SynthChannel[chn].env_arp->len) SynthChannel[chn].ptr_arp=SynthChannel[chn].env_arp->loop_ptr;

				if(!SynthChannel[chn].env_arp->abs[SynthChannel[chn].ptr_arp])
				{
					SynthChannel[chn].arp_offset=SynthChannel[chn].env_arp->value[SynthChannel[chn].ptr_arp++];
				}
				else
				{
					SynthChannel[chn].note=SynthChannel[chn].env_arp->value[SynthChannel[chn].ptr_arp++];
					SynthChannel[chn].arp_offset=0;
				}

				if(SynthChannel[chn].ptr_pitch>=SynthChannel[chn].env_pitch->len) SynthChannel[chn].ptr_pitch=SynthChannel[chn].env_pitch->loop_ptr;

				SynthChannel[chn].pitch_acc+=(float)SynthChannel[chn].env_pitch->value[SynthChannel[chn].ptr_pitch++];
			}

			//play channels

			sPITEnable=false;

			vol_max=-1;

			for(i=0;i<SYNTH_CHANNELS;++i)
			{
				chn=SynthChannelOrder[i];

				if(SynthChannel[chn].note<0) continue;

				if(!SynthChannel[chn].vol_output) continue;

				if(SynthChannel[chn].vol_output<vol_max) continue;

				vol_max=SynthChannel[chn].vol_output;

				freq=(float)(SynthChannel[chn].note-69);

				freq+=MidiPitchBend[SynthChannel[chn].midi_ch];

				freq+=SynthChannel[chn].arp_offset;

				freq+=SynthChannel[chn].pitch_acc/100.0f;

				sPITReload=(int)(1193180.0f/(440.0f*pow(2.0f,freq/12.0f)));

				sPITEnable=true;
			}

			if(sPITReload<1||sPITReload>65535)
			{
				if(sPITReload<1) sPITReload=1;
				if(sPITReload>65535) sPITReload=65535;

				sPITEnable=false;
			}

			LogFrame();
		}

		//generate sound

		level=0;
		loops=0;
		add=(int)sample_rate;

		while(sPITResampleAcc<1193180)
		{
			if(sPITEnable&&sPITOutput) level+=1.0f;

			sPITCount-=2;

			if(sPITCount<0)
			{
				sPITCount+=sPITReload;
				sPITOutput^=1;
			}

			sPITResampleAcc+=add;

			loops+=1.0f;
		}

		sPITResampleAcc-=1193180;

		level/=loops;

		level*=pOutputGain;

		(*outL++)=level;
		(*outR++)=level;
	}

	MidiQueue.clear();

	if(UpdateGuiFlag)
	{
		UpdateGUI(true);

		UpdateGuiFlag=false;
	}
}
